import json


def get_config_yaml():
    # read the file inferless.yaml in the current directory as a json string
    try:
        with open("inferless.yaml", "r") as file:
            config_yaml = file.read()
            return config_yaml
    except FileNotFoundError:
        raise Exception("Configuration file inferless.yaml not found in the current directory")
    except Exception as e:
        raise Exception(f"Error reading inferless.yaml file: {e}")



import numpy as np

np_map = {
    np.dtype('uint8'): "UINT8",
    np.dtype('uint16'): "UINT16",
    np.dtype('uint32'): "UINT32",
    np.dtype('uint64'): "UINT64",
    np.dtype('int8'): "INT8",
    np.dtype('int16'): "INT16",
    np.dtype('int32'): "INT32",
    np.dtype('int64'): "INT64",
    np.dtype('float16'): "FP16",
    np.dtype('float32'): "FP32",
    np.dtype('float64'): "FP64"
}


def create_data(prompts, is_batch=False):
    inputs = {
        "inputs": []
    }
    for key, value in prompts.items():
        data, data_type, shape = extract_data(value)
        if is_batch:
            shape = [1] + shape
        inputs["inputs"].append(
            {
                "data": data,
                "name": key,
                "shape": shape,
                "datatype": data_type,
            }
        )

    return inputs


def extract_data(data):
    shape = []
    item = data
    if isinstance(item, np.ndarray):
        shape = list(item.shape)
        data = item.tolist()
        data_type = np_map.get(item.dtype)
        return data, data_type, shape
    elif isinstance(item, list):
        while isinstance(item, list):
            shape.append(len(item))
            item = item[0]
        if isinstance(item, str):
            data_type = "BYTES"
        elif isinstance(item, int):
            data_type = "INT64"
        elif isinstance(item, float):
            data_type = "FP64"
        elif isinstance(item, bool):
            data_type = "BOOL"
        else:
            raise ValueError("Data type not supported, supported types are str, int, float, bool. Use data instead of "
                             "prompts to provide custom data")
        return data, data_type, shape
    elif isinstance(item, str):
        return [item], "BYTES", [1]
    elif isinstance(item, int):
        return [item], "INT64", [1]
    elif isinstance(item, float):
        return [item], "FP64", [1]
    elif isinstance(item, bool):
        return [item], "BOOL", [1]
    else:
        raise ValueError("Data type not supported, supported types are list, np array, str, int, float, bool. "
                         "Use inputs instead of data to provide custom inputs")


import os


def auth_header():
    try:
        access_token = os.getenv("INFERLESS_ACCESS_TOKEN")
        token_header = {"Authorization": f"Bearer {access_token}"}
        return token_header
    except Exception as e:
        print("Unable to fetch the credentials. Please login with inferless-cli. Follow instructions at "
              "https://docs.inferless.com/model-import/cli-import")
        raise SystemExit



import os
from typing import Optional, Any
import json
import threading
import requests
from functools import wraps
from .utils import create_data
from .rpc import get_rpc_payload, get_rpc_headers, get_rpc_result, get_rpc_url


def call(url: str, workspace_api_key: Optional[str] = None, data: Optional[dict] = None, callback: Optional[Any] = None,
         inputs: Optional[dict] = None, is_batch: Optional[bool] = False):
    """
    Call Inferless API
    :param url: Inferless Model API URL
    :param workspace_api_key: Inferless Workspace API Key
    :param data: Model Input Data as a dictionary, example: {"question": "What is the capital of France?", "context": "Paris is the capital of France."}
    :param callback: Callback function to be called after the response is received
    :param inputs: Model Input Data in inferless format
    :param is_batch: Whether the input is a batch of inputs, default is False
    :return: Response from the API call
    """
    try:
        if inputs is not None and data is not None:
            raise Exception("Cannot provide both data and inputs")

        if data is not None:
            inputs = create_data(data, is_batch)

        import requests
        if workspace_api_key is None:
            workspace_api_key = os.environ.get("INFERLESS_API_KEY")
        headers = {"Content-Type": "application/json",
                   "Authorization": f"Bearer {workspace_api_key}"}
        if inputs is None:
            inputs = {}
        response = requests.post(url, data=json.dumps(inputs), headers=headers)
        if response.status_code != 200:
            raise Exception(
                f"Failed to call {url} with status code {response.status_code} and response {response.text}")
        if callback is not None:
            callback(None, response.json())
        return response.json()
    except Exception as e:
        if callback is not None:
            callback(e, None)
        else:
            raise e


def call_async(url: str, workspace_api_key: Optional[str] = None, data: Optional[dict] = None,
               callback: Any = None, inputs: Optional[dict] = None, is_batch: Optional[bool] = False):
    """
    Call Inferless API
    :param url: Inferless Model API URL
    :param workspace_api_key: Inferless Workspace API Key
    :param data: Model Input Data as a dictionary, example: {"question": "What is the capital of France?", "context": "Paris is the capital of France."}
    :param callback: Callback function to be called after the response is received
    :param inputs: Model Input Data in inferless format
    :param is_batch: Whether the input is a batch of inputs, default is False
    :return: Response from the API call
    """
    thread = threading.Thread(target=call, args=(url, workspace_api_key, data, callback, inputs, is_batch))
    thread.start()
    return thread


def method(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        payload = get_rpc_payload(func, *args, **kwargs)
        headers = get_rpc_headers()
        url = get_rpc_url()
        response = requests.post(url, headers=headers, json=payload)
        return get_rpc_result(response)

    return wrapper



import base64
import json
import os

import dill
from . import config_yaml
from inferless.auth.token import auth_header


def get_rpc_payload(func, *args, **kwargs):
    rpc_payload = {
        "func": func,
        "args": args,
        "kwargs": kwargs
    }

    serialized_rpc_payload = base64.b64encode(dill.dumps(rpc_payload, recurse=True)).decode("utf-8")
    configuration_yaml = config_yaml.get_config_yaml()
    payload = {
        "rpc_payload": serialized_rpc_payload,
        "configuration_yaml": configuration_yaml
    }
    return payload


def get_rpc_headers():
    token_header = auth_header()
    headers = token_header.update(
        {
            "Content-Type": "application/json",
            "Accept": "application/json"
        }
    )
    return headers


def get_rpc_result(response):
    if response.status_code != 200:
        raise Exception(
            f"Failed to call rpc: status code {response.status_code} and response {response.text}")

    data = response.json()
    request_id = data.get("request_id")
    result = data.get("result")
    try:
        output = json.loads(base64.b64decode(base64.b64decode(result)))
        return output
    except Exception:
        raise Exception(f"Internal error occurred. Error message: {result}, Request ID for reference: {request_id}")


def get_rpc_url():
    if os.getenv("INFERLESS_ENV") == "dev":
        return "http://aab1b24401e6d40ee819a4a85da88501-394555867.us-east-1.elb.amazonaws.com/api/v1/rpc/start"

    return "https://serverless-region-v1.inferless.com/api/v1/rpc/start"
